/**
 * This interface describes a combo.
 * @author Christian Mattar
 *
 */

public class Combo {

	/**
	 * Bitfield to store buttons needed for transition
	 * The dancing mat is defined as a set of buttons:
     * 
     * Sl        St
     * 5    6    7
     * 3         4
     * 0    1    2
     * 
     * (Select and Start are not used here, they are just for orientation)
     * 
     **/
	public int keys[];
	
	/**
	 * Specify how long to wait minimally since arriving at current state
	 * until check for possible transition.
	 * 
	 */
	private int t_min[];

	/**
	 * Specify how long to wait maximally since arriving at current state
	 * until check for possible transition.
	 * 
	 */
	private int t_max[];

	/**
	 * Current state of the combo
	 */
	private int currentState;
	
	/**
	 * The next input has to occur after that
	 */
	private long nextStateMinTime;
	
	/**
	 * The next input has to occur before this
	 */
	private long nextStateMaxTime;
	
	
	Combo(int[] keys, int t_min[], int t_max[]) {
		if (keys.length != t_min.length ||
			keys.length != t_max.length) {
			throw new IllegalArgumentException("Combo data inconsistent!");
		}
		this.keys = new int[keys.length];
		this.t_min = new int[t_min.length];
		this.t_max = new int[t_max.length];
		System.arraycopy(keys,0,this.keys,0,keys.length);
		System.arraycopy(t_min,0,this.t_min,0,t_min.length);
		System.arraycopy(t_max,0,this.t_max,0,t_max.length);
	}
	
	public final int getCurrentState() {
		return currentState;
	}

	/**
	 * Returns whether a transition is not yet possible
	 * @param currentTimeSlice The current time
	 * @return True, if transition is not yet possible
	 */
	public final boolean isBeforeAdvanceWindow(long currentTimeSlice) {
		return currentTimeSlice < nextStateMinTime;
	}

	/**
	 * Returns whether a transition is not possible anymore
	 * @param currentTimeSlice The current time
	 * @return True, if transition is not possible anymore
	 */
	public final boolean missedAdvanceWindow(long currentTimeSlice) {
		return currentTimeSlice > nextStateMaxTime;
	}


	/**
	 * Transition to the next state
	 * Sets nextStateMinTime and nextStateMaxTime accordingly.
	 * @param currentTimeSlice The timeslice during which this
	 * transition occured.
	 */
	public final void advanceState(long currentTimeSlice) {
		currentState++;
		nextStateMinTime = t_min[currentState] + currentTimeSlice;
		nextStateMaxTime = t_max[currentState] + currentTimeSlice;
		
	}
	
	/**
	 * Reset this machine to the starting state
	 *
	 */
	public final void resetState() {
		currentState = 0;
		nextStateMinTime = 0;
		nextStateMaxTime = Long.MAX_VALUE;
	}
	
	/**
	 * Checks whether this combo is done
	 * @return True, if combo is done
	 */
	public final boolean isInFinalState() {
		if (currentState == keys.length -1) {
			return true;
		} else {
			return false;
		}
	}
}
