/**
 * Introducing: The Deterministic Finite Kunz Automaton
 * 
 *   -- dedicated to my good pal Georg Kunz --
 *
 * This class detects movement-combos on the dance mat.
 * Each combo is basically a state machine, where a transition from state
 * q to q+1 occurs iff button b_n is pressed and 
 * t_min_q < t < t_max_q where t is time since reaching state q.
 * If b_n is not pressed in that interval,
 * the state machine resets to its starting state.
 * Should the machine reach its final state (always its last
 * state), a combo is completed.
 * 
 * Since we have multiple different combos, we basically simulate a
 * product automata, with the data for the different combos saved in arrays.
 * We only have to look at the data from the most current timeslice to
 * decide whether transitions occur.
 * So if we manage m combos, we can do this very efficiently in O(m).
 * 
 */

import java.util.Vector;

public class ComboManager {
	
	Vector combos;
	
	ComboManager() {
		combos = new Vector();
	}
	
	/**
	 * Adds a new combo to our machine.
	 * Note: Deleting a combo is impossible right now, since it would
	 * lead to wrong reference indexes. It shouldn't be necessary anyway.
	 * @param combo Combo to add
	 * @return Index of the newly added combo
	 */
	public int registerCombo(Combo combo) {
		combo.resetState();
		combos.add(combo);
		return combos.size();
	}
	
	/**
	 * Reset all managed combos to their starting states
	 */
	public void resetAllCombos() {
		Combo combo;
		for(int i=0;i<combos.size();i++) {
			combo = (Combo) combos.get(i);
			combo.resetState();
		}
	}
	
	/**
	 * Should be called once per timeslice.
	 * @param currentTimeSlice The current timeslice index
	 * @return If a combo occured, the index of the combo.
	 * Else -1
	 * 
	 */
	public int checkForCombos(int keys, long currentTimeSlice) {
		int size = combos.size();
		Combo currentCombo;
		int currentState;

		for(int i=0;i<combos.size();i++) {
			currentCombo = (Combo) combos.get(i);
			currentState = currentCombo.getCurrentState();
			
			// If we missed our window for advancing to the next
			// state, return to starting state and continue with
			// next machine
			if (currentCombo.missedAdvanceWindow(currentTimeSlice)) {
				System.out.println("Missed Combo. Resetting...");
				currentCombo.resetState();
				continue;
			}
			
			// If our opportunity for advancing to the next state
			// hasn't arrived yet, continue with next machine
			if (currentCombo.isBeforeAdvanceWindow(currentTimeSlice)) {
				continue;
			}
			
			// Check if any of the played notes is needed to
			// advance in state
			if ((keys & currentCombo.keys[currentState]) != 0) {
				currentCombo.advanceState(currentTimeSlice);
				System.out.println("Advanced State to" + currentCombo.getCurrentState());
			}
				
			// Check whether we've completed the combo
			if (currentCombo.isInFinalState()) {
				// If we do, reset all combos and return the index
				// of the performed combo
				resetAllCombos();
				return i;
			}
		}
		
		return -1;
	}

}
