import java.awt.Component;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.BitSet;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Calendar;

/**
 * This class implements a Dance instrument.
 * 
 * It provides a mechanism to process input and react in a 
 * musical way by playing rhythm, notes, and detecting freely
 * definable combos.
 * 
 * You have to use the |run| method to initialize the timer thread,
 * and the stop method to tear it down.
 * 
 */
public class Dance implements KeyListener{


	private static final long serialVersionUID = 3034621072560498996L;
	public final static int TIMESLICEINTERVAL = 80;

	public final static int BUTTON_BL = 0;
	public final static int BUTTON_B = 1;
	public final static int BUTTON_BR = 2;
	public final static int BUTTON_L = 3;
	public final static int BUTTON_R = 4;
	public final static int BUTTON_TL = 5;
	public final static int BUTTON_T = 6;
	public final static int BUTTON_TR = 7;
	public final static int BUTTON_SEL = 8;

	public final static int KEY_ERROR = -1;
	public final static int KEY_C = 0;
	public final static int KEY_D = 1;
	public final static int KEY_E = 2;
	public final static int KEY_F = 3;
	public final static int KEY_G = 4;
	public final static int KEY_A = 5;
	public final static int KEY_H = 6;
	
	// Ugly hack:
	// Save the keys we played whi
	private BitSet ChordNote;
	
	public final static int CHORDS[] = {
			(int) (Math.pow(2,KEY_C)+Math.pow(2,KEY_E)),
			(int) (Math.pow(2,KEY_D)+Math.pow(2,KEY_A)),
			(int) (Math.pow(2,KEY_E)+Math.pow(2,KEY_H)),
			(int) (Math.pow(2,KEY_F)+Math.pow(2,KEY_A)),
			(int) (Math.pow(2,KEY_G)+Math.pow(2,KEY_H)),
			(int) (Math.pow(2,KEY_A)+Math.pow(2,KEY_E)),
			(int) (Math.pow(2,KEY_D)+Math.pow(2,KEY_F)),
	};
	
	/**
	 * Bitfield which stores which buttons were pressed in the current
	 * timeslice
	 * Note: We use int here instead of BitSet, because we'll never have more than
	 * 32 buttons anyway, and the additional BitSet-API handling would make things
	 * quite more complicated
	 */
	private int buttonsPressedField;

	/**
	 * Bitfield which stores which keys were played last frame, so that we
	 * don't play them again without releasing the button first.
	 * Note: We use int here instead of BitSet, because we'll never have more than
	 * 32 buttons anyway, and the additional BitSet-API handling would make things
	 * quite more complicated
	 */
	private int previousButtonPressedField;

	
	/**
	 * Stores which rhythm is currently playing
	 */
	private int currentRhythm;

	public Timer advanceTimeSliceTask;
	
	/**
	 * Stores the current timeslice. Gets incremented by
	 * advanceTimeSliceTask
	 */
	private int currentTimeSlice;
	
	/**
	 * Pointer to output mechanism
	 */
	private DanceOutput output;

	/**
	 * Pointer to ComboManager
	 */
	private ComboManager comboManager;
	
	/**
	 * Rhythm mode or compose mode
	 */
	private boolean selectState;
	
	private int currentChordKeys;
	
	/**
	 * This class gets called at the end of each time slice to
	 * process the input.
	 */
	private class AdvanceTimeSlice extends TimerTask {

		public synchronized void run() {
			int keys;
			int cleanedUpKeys;
			
			// Remove the keys that were played previous time slice from the bitfield
			// This should prevent stuttering
			cleanedUpKeys = buttonsPressedField & ~previousButtonPressedField;

			// If a button is released, stop its playback
			stopKeys(buttonsPressedField, previousButtonPressedField);

			
			previousButtonPressedField = buttonsPressedField;
			
			// If select key is pressed, change mode and forget about the other
			// pressed buttons
			if ((cleanedUpKeys & (int) Math.pow(2,BUTTON_SEL)) != 0) {
				selectState = !selectState;
				return;
			}
			
			// In compose mode, play the pressed keys and check for combos
			if (selectState == false) {
				// If keys were already played as a chord, don't play them again as single notes.
				if (!handleChords(cleanedUpKeys)) {
					playKeys(cleanedUpKeys);
				}
				int combo = comboManager.checkForCombos(cleanedUpKeys, currentTimeSlice);
				if (combo != -1) {
					// Increasing actual output by 1, because it's
					// easier to do in Max/MSP
					output.playCombo(combo+1);
				}
				
			// In rhythm mode, see if we have to change/stop the current rhythm
			} else if (selectState == true) {
				handleRhythm(cleanedUpKeys);
			}

			// Clear button state
			buttonsPressedField = 0;
			currentTimeSlice++;
		}
		
	}
	
	/**
	 * Takes care of playing chords: Certain key combinations lead to
	 * playing a chord.
	 * @param keys Keys being pressed
	 * @return Returns true if a chord was played
	 */
	private boolean handleChords(int keys) {
		for(int i=0;i<CHORDS.length;i++) {
			if (keys == CHORDS[i]) {
				// Increase actual output by one, since that
				// is easier to handle in Max/MSP
				output.playChord(i+1);

				// Save the keys we used to play the chord so that on release,
				// they don't stop playing.
				currentChordKeys = CHORDS[i];
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * 
	 * @param keys Bitfield containing key state
	 */
	private void handleRhythm(int keys) {

		int rhythm;
		
		// If multiple buttons are pressed, just take the first one, and forget
		// about the rest
		for(int i=0;keys != 0;i++) {
			if ((keys & 1) != 0) {
				if (currentRhythm == i) {
					stopRhythm();
				} else {
					startRhythm(i);
				}
				return;
			}
			keys = keys >> 1;
		}
		
		
		
	}
	
	/**
	 * Constructor method.
	 * Use the |run| method to actually start the AdvanceTimeSliceTask
	 * @param output Output mechanism
	 * @param inputComponent AWT Component at which we should listen for
	 * input events
	 */
	public Dance(DanceOutput output, Component inputComponent) {
		this.output = output;
		currentRhythm = -1;
		currentTimeSlice = 0;
		buttonsPressedField = 0;
		comboManager = new ComboManager();
		for(int i=0;i<ComboStorage.length();i++) {
			comboManager.registerCombo(ComboStorage.getCombo(i));
		}
		inputComponent.addKeyListener(this);
		advanceTimeSliceTask = new Timer();
	}
	
	/**
	 * Play the keys marked in the bitfield
	 * @param keys Bitfield of key states
	 */
	public void playKeys(int keys) {
		for(int i=0;keys != 0;i++) {
			if ((keys & 1) != 0) {
				// Increase actual key output by one, since that
				// is easier to handle in Max/MSP
				output.playKey(i+1);
			}
			keys = keys >> 1;
		}
		return;
	}
	
	/**
	 * Stop the midioutput for all keys which are pressed in oldField but
	 * not in newField
	 * @param newField Bitfield of key states
	 * @param oldField Bitfield of key states
	 */
	public void stopKeys(int newField, int oldField) {
		int currentChordKeysTemp = currentChordKeys;
		for(int i=0; oldField != 0; i++) {
			if ((oldField & 1) != 0 && (newField & 1) == 0) {
				if ((currentChordKeysTemp & 1) == 0) {
					output.stopKey(i+1);
				} else {
					// If the key we are supposed to stop was used to
					// play a chord, don't stop it but clear the flag instead
					currentChordKeys = currentChordKeys & ~((int) Math.pow(2,i));
				}
			}
			newField = newField >> 1;
			oldField = oldField >> 1;
			currentChordKeysTemp = currentChordKeysTemp >> 1;
		}
	}
	
	/**
	 * Stop playing the background rhythm
	 *
	 */
	public void stopRhythm() {
		currentRhythm = -1;
		output.stopRhythm();
	}
	
	/**
	 * Start the Timer thread to process the input
	 *
	 */
	public void run() {
		advanceTimeSliceTask.schedule(new AdvanceTimeSlice(),
				Calendar.getInstance().getTime(), TIMESLICEINTERVAL);
	}

	/**
	 * Necessary clean-up of threads
	 *
	 */
	public void stop() {
		advanceTimeSliceTask.cancel();
	}

	/**
	 * Start playing a new background rhythm
	 * @param rhythm
	 */
	public void startRhythm(int rhythm) {
		currentRhythm = rhythm;
		
		// Increasing actual output by 1, because it's
		// easier to do in Max/MSP
		output.startRhythm(rhythm+1);
	}
	
	
	/**
	 * Check for keys pressed
	 * Unfortunately, due to driver problems, we are using an ugly hack driver which
	 * translates the dance mats output to keyboard input. Unfortunately, we don't get
	 * true button press/release information there, so we have to work around this.
	 * 
	 * For easier emulation of the mat on the keyboard, we use
	 * the num pad-arrows as buttons, e.g. 9 represents the upper right
	 * arrow, and 5 is not used at all.
	 * Internally, we translate those to 0-7:
	 * 
     * /         St
     * 5    6    7
     * 3         4
     * 0    1    2
     * 
     * (Start is not used here, it's just for orientation)
	 * 
	 * We save the button states by using an integer as a bitfield
	 * 
	 */
	public synchronized void keyPressed(KeyEvent e) {
		int button;
		
		if (e.getKeyChar() == '/') {
			button = BUTTON_SEL;
		} else if (e.getKeyChar() == '1') {
			button = BUTTON_BL;
		} else if (e.getKeyChar() == '2') {
			button = BUTTON_B;
		} else if (e.getKeyChar() == '3') {
			button = BUTTON_BR;
		} else if (e.getKeyChar() == '4') {
			button = BUTTON_L;
		} else if (e.getKeyChar() == '6') {
			button = BUTTON_R;
		} else if (e.getKeyChar() == '7') {
			button = BUTTON_TL;
		} else if (e.getKeyChar() == '8') {
			button = BUTTON_T;
		} else if (e.getKeyChar() == '9') {
			button = BUTTON_TR;
		} else {
			// Throw away any other key
			return;
		}

		buttonsPressedField |= (int) Math.pow(2,button);
	}

	/*
	 * stub needed to implement KeyListener
	 */
	public void keyTyped(KeyEvent e) {
	}

	/*
	 * stub needed to implement KeyListener
	 */
	public void keyReleased(KeyEvent e) {
	}
	

}
